/**
 * This is part of an XML schema Apache module.
 * 
 * Copyright (C) 2006 Nokia Corporation.
 * 
 * Contact: Jari Urpalainen <jari.urpalainen@nokia.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 * $Date:$
 */

#if 0
LoadModule schema_module modules/mod_schema.so

<Location /schema>
   SetHandler schema_handler
   SchemaProcess DOMPrint
   HttpProxy http://example.com:8080
</Location>
#endif

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_log.h"
#include "http_main.h"
#include "http_core.h"
#include "ap_config.h"
#include "apreq.h"
#include "apreq_module.h"
#include "apr_strings.h"
#include "apreq2/apreq_module_apache2.h"

#include <libxml/tree.h>
#include <libxml/xpath.h>

#include <unistd.h>
#include <glib.h>

module AP_MODULE_DECLARE_DATA schema_module;

/* configuration data */
typedef struct _schema_cfg {
  char *pchSchemaProcess,
       *pchHttpProxy; 
} schema_cfg;

/* parsing multipart */
typedef struct _schema_context 
{
  struct _schema_context *next;
  
  request_rec *r;
  xmlDocPtr doc;
  char *pchFileName,
       *pchLocalFileName, 
       *pchTargetNamespace;
  int fSchema, fError, fReferenced, vlen;
  apr_pool_t *pool;

} schema_context_t;  
   
/* create local temporary files */
static 
apr_file_t *create_schema_file_name(apr_pool_t *pool, const char *tn, char **pp)
{
  apr_file_t *f;
  char *dirname;

  if (apr_temp_dir_get((const char**)&dirname, pool) != APR_SUCCESS) 
    return NULL;
    
  *pp = dirname = (char*)apr_psprintf(pool, "%s/%s", dirname, tn);

  if (apr_file_mktemp(&f, dirname, APR_CREATE | APR_READ | APR_WRITE | APR_EXCL, pool) != APR_SUCCESS)
    return NULL; 
  return f;
}   

/** simple full xpath query over doc */
static
void process_xpath(xmlDocPtr doc,  
                   const char *pcsz,
                   xmlNodeSetPtr pNodeSet)
{
  xmlXPathCompExprPtr comp = xmlXPathCompile((const xmlChar*)pcsz);
  xmlNodePtr node = xmlDocGetRootElement(doc);
                       
  if (comp) {
    xmlXPathObjectPtr res = NULL;
    xmlXPathContextPtr ctxt = NULL;
    
    ctxt = xmlXPathNewContext(doc);
    ctxt->node = node;
    res = xmlXPathCompiledEval(comp, ctxt);
    xmlXPathFreeCompExpr(comp);
    
    if (res && res->nodesetval && res->nodesetval->nodeNr) {
      int cb;  

      pNodeSet->nodeNr = res->nodesetval->nodeNr;
      pNodeSet->nodeTab = malloc(cb = pNodeSet->nodeNr * sizeof(xmlNodePtr));
      memcpy(pNodeSet->nodeTab, res->nodesetval->nodeTab, cb);  

      xmlXPathFreeObject(res); 
      xmlXPathFreeContext(ctxt); 
    } else {
      pNodeSet->nodeNr = 0;
      pNodeSet->nodeTab = NULL;
    }
  } else {
    pNodeSet->nodeNr = 0;
    pNodeSet->nodeTab = NULL;
  }           
  return;
}

/* free xpath results */
static
void free_xpath(xmlNodeSetPtr pNodeSet)
{
  if (pNodeSet) {
    free(pNodeSet->nodeTab);
    pNodeSet->nodeNr = 0;
  }  
}

/* parse params & files */
static 
int schema_dump_table(void *ctx, const char *key, const char *value)
{
  schema_context_t *schema = ctx;
  request_rec *r = schema->r;
  apreq_param_t *param = apreq_value_to_param(value);

  if (param && param->upload) {
    char *pch, *pchTemp;
    apr_size_t vlen;
    schema_context_t *s;
    xmlNodeSet aNodeSet[1];
    apr_file_t *fp;
 
    apr_brigade_pflatten(param->upload, &pch, &vlen, schema->pool);
    if (!vlen)  /* an error case */
      return  1;
  
    for(s = schema; s && s->next; s = s->next)
      ;

    if (s->doc || s->vlen) {
      s->next = apr_pcalloc(schema->pool, sizeof(*s));
      
      s = s->next;
      s->r = r;
    }
    xmlKeepBlanksDefault(0);
    s->doc = xmlParseMemory(pch, vlen);
    s->vlen = vlen;
    
    s->fSchema = strcmp(key, "instance");

    s->pchFileName = (char*)apr_psprintf(schema->pool, "%s", value);
    pchTemp = (char*)apr_psprintf(schema->pool, "%s.XXXXXX", s->pchFileName);

    fp = create_schema_file_name(schema->pool, 
                                 pchTemp, 
                                 &s->pchLocalFileName);
    if (s->doc) {
      process_xpath(s->doc, "/*/@targetNamespace", aNodeSet);

      if (aNodeSet->nodeNr && aNodeSet->nodeTab[0]->children)
        s->pchTargetNamespace = 
             (char*)apr_psprintf(schema->pool, "%s", 
                                 aNodeSet->nodeTab[0]->children->content);
      free_xpath(aNodeSet);  
    }  
    /* store received file as-is */
    if (fp) {
      apr_size_t cb; 
        
      apr_file_write_full(fp, (const char*)pch, vlen, &cb);            
      apr_file_close(fp);
    }
  }
  return 1;
}

/** dump files with error references */
static
void dump_schemas(schema_context_t *s)
{
  for(; s; s = s->next) {
    if (s->fError) {
      xmlChar *pch = NULL;
      int cb; 

      xmlKeepBlanksDefault(0);
      xmlDocDumpFormatMemoryEnc(s->doc, &pch, &cb, "UTF-8", 1);

      ap_rprintf(s->r, 
                 "\n%s%s:\n%s", 
                 s->fSchema ? "Schema " : "Instance ", 
                 s->pchFileName, 
                 pch ? (char*)pch : "ERROR: could not be parsed.\n"); 
      xmlFree(pch);
    }
  }
}

/* change local file references */
static
gchar *schema_replace_str(char *pchErr, 
                          const char *pchS, 
                          const char *pchR, 
                          int *f)
{
  char *pR = pchErr;
  int s, r, c;

  *f = FALSE;
  if (pchS == NULL || pchR == NULL)
    return pchErr;

  if ((s = strlen(pchS)) < (r = strlen(pchR)))
    return pchErr;

  for(;;) {
    pR = strstr(pR, pchS);
    if (pR) {
      memcpy(pR, pchR, r);
      memmove(pR + r, pR + s, c = strlen(pR + s));
      pR[c + r] = 0;
      pR += r; 
      *f = TRUE;
    } else
      break;
  }
  return pchErr;  
}

/* dump parser errors */
static 
void schema_error(void *ctx, const char *fmt, ...) 
{
  schema_context_t *s = ctx;
  int fError;
  char *pchErr;
  va_list va;
 
  if (s == NULL)
    return;   
 
  va_start(va, fmt);
  pchErr = apr_pvsprintf(s->r->pool, fmt, va);
  va_end(va);
    
  schema_replace_str(pchErr, s->pchLocalFileName, s->pchFileName, &fError);
  ap_rprintf(s->r, "%s", pchErr); 
}

/* schema handler callback */
static 
int schema_handler(request_rec *r)
{
  apreq_handle_t *req;
  const apr_table_t *t;
  schema_context_t schema[1], *s, *sInstance = NULL;
  schema_cfg *scfg = (schema_cfg *)
             ap_get_module_config(r->server->module_config, 
                                  &schema_module);

  if (strcmp(r->handler, "schema_handler") != 0)
    return DECLINED;

  memset(schema, 0, sizeof(*schema));
  req = (apreq_handle_t *)apreq_handle_apache2(r);

  schema->r = r; 
  apr_pool_create(&schema->pool, NULL);
    
  ap_set_content_type(r, "text/plain");

  if (apreq_args(req, &t) == APR_SUCCESS)
    apr_table_do(schema_dump_table, schema, t, NULL);

  if (apreq_body(req, &t) == APR_SUCCESS) 
    apr_table_do(schema_dump_table, schema, t, NULL);

  ap_rprintf(r, "Validation results\n\n"
                "Checking first import/include references in schema files:\n");

  for(s = schema; s; s = s->next) {
    int fOk = TRUE;

    if (!s->pchFileName)
      continue;
    
    ap_rprintf(r, "%s: ", s->pchFileName);
  
    if (s->doc) {
      xmlNodeSet aNodeSet[1];

      if (s->fSchema) {
        int i;   
        process_xpath(s->doc, "//@schemaLocation", aNodeSet);

        for(i = 0; i < aNodeSet->nodeNr; i++) {
          schema_context_t *t;
          const char *pR, *pcsz = (const char*)
                                  aNodeSet->nodeTab[i]->children->content;

          pR = strrchr(pcsz, '/');   
          if (pR) 
            pcsz = pR + 1;

          for(t = schema; t; t = t->next) {
            if (!strcmp(pcsz, t->pchFileName)) {
              /* change the reference to the submitted local file */
              xmlNodeSetContent(aNodeSet->nodeTab[i], 
                                (const xmlChar*)t->pchLocalFileName);
              t->fReferenced = TRUE;
              
              /* remember to also store the changed file */
              xmlSaveFile(s->pchLocalFileName, s->doc);
              break;
            }   
          } 
          if (t == NULL) {
            xmlNodePtr node = aNodeSet->nodeTab[i]->parent;
            xmlChar *attr = xmlGetProp(node, (const xmlChar*)"namespace");
            if (!attr || strcmp((const char*)attr, 
                                "http://www.w3.org/XML/1998/namespace")) {
              ap_rprintf(r, "\n\tERROR: schema reference '%s' not found !"
                            " Submit that schema too.", pcsz);
              fOk = FALSE;
            }  
            xmlFree(attr);             
          }             
        } 
        free_xpath(aNodeSet);

      } else {
        int i;   
        
        process_xpath(s->doc, "//@*[local-name()=\"schemaLocation\"]"
                              "[namespace-uri()=\"http://www.w3.org/2001/XMLSchema-instance\"]", 
                              aNodeSet);

        for(i = 0; i < aNodeSet->nodeNr; i++) {
          ap_rprintf(r, "\n\tWARNING: schemaLocation hint removed from the instance document %s", 
                        s->pchFileName);

          xmlRemoveProp((xmlAttrPtr)aNodeSet->nodeTab[i]);
          fOk = FALSE;
        }
        free_xpath(aNodeSet);
        sInstance = s; 
      }   
   
    } else if (s->vlen) {
      ap_rprintf(r, "ERROR: Could not parse the XML file. Check that it is XML well-formed !\n");
   
      xmlSetGenericErrorFunc(s, schema_error);
   
      xmlFreeDoc(xmlReadFile(s->pchLocalFileName, NULL, XML_PARSE_PEDANTIC));
      
      xmlSetGenericErrorFunc(NULL, NULL);
    }
   
    if (s->doc)
      ap_rprintf(r, "%s\n", fOk ? "OK" : "");
  }

  { /* xerces c++ w3c schema validation tool needs some command line parameters */
    char *pchSchema = NULL, 
         *pchNoNsSchema = NULL,
         *pchOut, 
         *pchErr, 
         *pchHttpProxy = scfg && scfg->pchHttpProxy ? 
                         (char*)apr_psprintf(schema->pool, 
                                             "http_proxy=\"%s\"", 
                                             scfg->pchHttpProxy) : NULL;
                                          
    GError *error;
    gint rc, i = 5;
    char *argv[] = { scfg && scfg->pchSchemaProcess ? scfg->pchSchemaProcess : "DOMPrint",
                     "-f", "-s", "-n", "-v=always", NULL, NULL, NULL, NULL, NULL },
         *arge[] = { pchHttpProxy, NULL };      

    for(s = schema; s; s = s->next) {
      if (s->fSchema && !s->fReferenced) {
        if (s->pchTargetNamespace) 
          pchSchema = (char*)apr_psprintf(schema->pool, "%s%s%s %s", 
                                          pchSchema ? pchSchema : "-schema=",
                                          pchSchema ? " " : "",
                                          s->pchTargetNamespace, s->pchLocalFileName);
        else 
          pchNoNsSchema = (char*)apr_psprintf(schema->pool, "%s%s%s", 
                                              pchNoNsSchema ? pchNoNsSchema : "-nonsschema=",
                                              pchNoNsSchema ? " " : "",
                                              s->pchLocalFileName);
      }                                    
    }
    if (!pchSchema && !pchNoNsSchema) 
      ap_rprintf(r, "\nFailed to run validation process.\n"
                    "schema file(s) missing, you should upload at least one.\n");
    else if (!sInstance || !sInstance->pchLocalFileName) 
      ap_rprintf(r, "\nFailed to run validation process.\n"
                    "An instance document missing, you should upload it too.\n");
    else {
      if (pchSchema)
        argv[i++] = pchSchema;
      if (pchNoNsSchema)
        argv[i++] = pchNoNsSchema;
      if (sInstance && sInstance->pchLocalFileName)   
        argv[i++] = sInstance->pchLocalFileName;
      argv[i++] = NULL;

      if (!g_spawn_sync(NULL,
                        argv,
                        arge, 
                        G_SPAWN_SEARCH_PATH,
                        NULL,
                        NULL,
                        &pchOut,
                        &pchErr, 
                        &rc, 
                        &error))
        ap_rprintf(r, "\nFailed to run validation process '%s',\n"
                      "configuration error in the server ?\n", 
                      argv[0]);
      else  {
        ap_rprintf(r, "\nRunning validation process '%s'\n", argv[0]);
        if (pchOut)
          ap_rprintf(r, "%s", pchOut); 

        if (pchErr && strlen(pchErr) != strspn(pchErr, " \t\n\r")) {
          ap_rprintf(r, "\nErrors detected:\n"); 

          for(s = schema; s; s = s->next) 
            schema_replace_str(pchErr, s->pchLocalFileName, s->pchFileName, &s->fError);
        
          ap_rprintf(r, "%s", pchErr); 
          dump_schemas(schema);
        }  
        g_free(pchOut);  
        g_free(pchErr);  
      }  
    }
  }
  /* cleanup */
  for(s = schema; s; s = s->next) {
    xmlFreeDoc(s->doc);
    s->doc = NULL;

    if (s->pchLocalFileName)
      unlink(s->pchLocalFileName); 
  }
  apr_pool_destroy(schema->pool);
  return OK;
}

/* server config create */
static void *schema_create_server_config(apr_pool_t *p, server_rec *s)
{
  return apr_pcalloc(p, sizeof(schema_cfg));
}

/* schema process directive */
static const char *cmd_schema_process(cmd_parms *cmd, 
                                      void *mconfig, 
                                      const char *pch)
{
  server_rec *s = cmd->server;

  schema_cfg *conf = (schema_cfg *)
                     ap_get_module_config(s->module_config, &schema_module);

  if (pch) 
    conf->pchSchemaProcess = (char*)apr_psprintf(cmd->pool, "%s", pch);
  return NULL;
}

/* http_proxy directive */
static const char *cmd_http_proxy(cmd_parms *cmd, 
                                  void *mconfig, 
                                  const char *pch)
{
  server_rec *s = cmd->server;

  schema_cfg *conf = (schema_cfg *)
                     ap_get_module_config(s->module_config, &schema_module);

  if (pch) 
    conf->pchHttpProxy = (char*)apr_psprintf(cmd->pool, "%s", pch);
  return NULL;
}

/* cmd callbacks */
static const command_rec schema_cmds[] =
{
    AP_INIT_TAKE1(
        "SchemaProcess",                    /* directive name */
        cmd_schema_process,                 /* config action routine */
        NULL,                               /* argument to include in call */
        OR_OPTIONS,                         /* where available */
        "W3C Schema validation process"     /* directive description */
      ),
    AP_INIT_TAKE1(
        "HttpProxy",                        /* directive name */
        cmd_http_proxy,                     /* config action routine */
        NULL,                               /* argument to include in call */
        OR_OPTIONS,                         /* where available */
        "http proxy for schema processing"  /* directive description */
      ),
    {NULL}
};

/* initialize hhoks */
static void register_hooks(apr_pool_t *p)
{
  ap_hook_handler(schema_handler, NULL, NULL, APR_HOOK_MIDDLE);
}

module AP_MODULE_DECLARE_DATA schema_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,     		/* per-directory config creator */
    NULL,     		/* dir config merger */
    schema_create_server_config, /* server config creator */
    NULL,     		/* server config merger */
    schema_cmds,     	/* command table */
    register_hooks,	/* set up other request processing hooks */
};

